@@ -564,7 +564,7 @@ radv_postconfig(struct proto_config *CF)
   // struct radv_config *cf = (void *) CF;
 
   /* Define default channel */
-  if (EMPTY_LIST(CF->channels))
+  if (! proto_cf_main_channel(CF))
     channel_config_new(NULL, net_label[NET_IP6], NET_IP6, CF);
 }
 
@@ -740,7 +740,7 @@ radv_pref_str(u32 pref)
 
 /* The buffer has some minimal size */
 static int
-radv_get_attr(eattr *a, byte *buf, int buflen UNUSED)
+radv_get_attr(const eattr *a, byte *buf, int buflen UNUSED)
 {
   switch (a->id)
   {
